# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: mir_controller_service.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from ymir_proto.gen import mir_common_pb2 as mir__common__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='mir_controller_service.proto',
  package='ymir.controller',
  syntax='proto3',
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\x1cmir_controller_service.proto\x12\x0fymir.controller\x1a\x10mir_common.proto\"\xa9\x03\n\nGeneralReq\x12\x0f\n\x07user_id\x18\x01 \x01(\t\x12\x0f\n\x07repo_id\x18\x02 \x01(\t\x12.\n\x08req_type\x18\x03 \x01(\x0e\x32\x1c.ymir.controller.RequestType\x12\x0f\n\x07task_id\x18\x04 \x01(\t\x12\x14\n\x0csingleton_op\x18\x05 \x01(\t\x12\x13\n\x0bhis_task_id\x18\x06 \x01(\t\x12\x13\n\x0b\x64st_task_id\x18\x07 \x01(\t\x12\x16\n\x0ein_dataset_ids\x18\x08 \x03(\t\x12\x16\n\x0e\x65x_dataset_ids\x18\t \x03(\t\x12\x14\n\x0cin_class_ids\x18\n \x03(\x05\x12\x14\n\x0c\x65x_class_ids\x18\x0b \x03(\x05\x12\r\n\x05\x66orce\x18\x0c \x01(\x08\x12\x16\n\x0e\x63ommit_message\x18\r \x01(\t\x12\x38\n\x0freq_create_task\x18\xe9\x07 \x01(\x0b\x32\x1e.ymir.controller.ReqCreateTask\x12;\n\x11req_get_task_info\x18\xea\x07 \x01(\x0b\x32\x1f.ymir.controller.ReqGetTaskInfo\"\x92\x01\n\x0bGeneralResp\x12\x0c\n\x04\x63ode\x18\x01 \x01(\x05\x12\x13\n\x0breq_task_id\x18\x02 \x01(\t\x12\x0f\n\x07message\x18\x03 \x01(\t\x12\x10\n\x08\x65xt_strs\x18\x04 \x03(\t\x12=\n\x12resp_get_task_info\x18\xe8\x07 \x01(\x0b\x32 .ymir.controller.RespGetTaskInfo\"\xcf\x03\n\rReqCreateTask\x12(\n\ttask_type\x18\x01 \x01(\x0e\x32\x15.ymir.common.TaskType\x12\x15\n\rsampling_rate\x18\x02 \x01(\x02\x12\x17\n\x0fno_task_monitor\x18\x03 \x01(\x08\x12.\n\x06\x66ilter\x18\x65 \x01(\x0b\x32\x1e.ymir.controller.TaskReqFilter\x12\x32\n\x08training\x18\x66 \x01(\x0b\x32 .ymir.controller.TaskReqTraining\x12.\n\x06mining\x18g \x01(\x0b\x32\x1e.ymir.controller.TaskReqMining\x12\x34\n\timporting\x18h \x01(\x0b\x32!.ymir.controller.TaskReqImporting\x12\x34\n\texporting\x18i \x01(\x0b\x32!.ymir.controller.TaskReqExporting\x12\x34\n\tinference\x18j \x01(\x0b\x32!.ymir.controller.TaskReqInference\x12.\n\x04\x63opy\x18k \x01(\x0b\x32 .ymir.controller.TaskReqCopyData\"S\n\rTaskReqFilter\x12\x16\n\x0ein_dataset_ids\x18\x01 \x03(\t\x12\x14\n\x0cin_class_ids\x18\x02 \x03(\x05\x12\x14\n\x0c\x65x_class_ids\x18\x03 \x03(\x05\"\xce\x01\n\x0fTaskReqTraining\x12N\n\x10in_dataset_types\x18\x01 \x03(\x0b\x32\x34.ymir.controller.TaskReqTraining.TrainingDatasetType\x12\x14\n\x0cin_class_ids\x18\x02 \x03(\x05\x1aU\n\x13TrainingDatasetType\x12\x12\n\ndataset_id\x18\x01 \x01(\t\x12*\n\x0c\x64\x61taset_type\x18\x02 \x01(\x0e\x32\x14.ymir.common.TvtType\"b\n\rTaskReqMining\x12\x16\n\x0ein_dataset_ids\x18\x01 \x03(\t\x12\x16\n\x0e\x65x_dataset_ids\x18\x02 \x03(\t\x12\x12\n\nmodel_hash\x18\x03 \x01(\t\x12\r\n\x05top_k\x18\x04 \x01(\x05\"=\n\x10TaskReqImporting\x12\x11\n\tasset_dir\x18\x01 \x01(\t\x12\x16\n\x0e\x61nnotation_dir\x18\x02 \x01(\t\"{\n\x10TaskReqExporting\x12\x12\n\ndataset_id\x18\x01 \x01(\t\x12(\n\x06\x66ormat\x18\x02 \x01(\x0e\x32\x18.ymir.common.LabelFormat\x12\x11\n\tasset_dir\x18\x03 \x01(\t\x12\x16\n\x0e\x61nnotation_dir\x18\x04 \x01(\t\"\x12\n\x10TaskReqInference\"S\n\x0fTaskReqCopyData\x12\x13\n\x0bsrc_user_id\x18\x01 \x01(\t\x12\x13\n\x0bsrc_repo_id\x18\x02 \x01(\t\x12\x16\n\x0esrc_dataset_id\x18\x03 \x01(\t\"\"\n\x0eReqGetTaskInfo\x12\x10\n\x08task_ids\x18\x01 \x03(\t\"\xa7\x01\n\x0fRespGetTaskInfo\x12\x43\n\ntask_infos\x18\x01 \x03(\x0b\x32/.ymir.controller.RespGetTaskInfo.TaskInfosEntry\x1aO\n\x0eTaskInfosEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12,\n\x05value\x18\x02 \x01(\x0b\x32\x1d.ymir.controller.TaskInfoItem:\x02\x38\x01\"\x95\x02\n\x0cTaskInfoItem\x12\x0f\n\x07task_id\x18\x01 \x01(\t\x12\x14\n\x0c\x62\x61se_task_id\x18\x02 \x01(\t\x12\x0f\n\x07user_id\x18\x03 \x01(\t\x12\x0f\n\x07repo_id\x18\x04 \x01(\t\x12#\n\x04type\x18\x05 \x01(\x0e\x32\x15.ymir.common.TaskType\x12%\n\x05state\x18\x06 \x01(\x0e\x32\x16.ymir.common.TaskState\x12\x17\n\x0fstart_timestamp\x18\x07 \x01(\x03\x12\x18\n\x10\x66inish_timestamp\x18\x08 \x01(\x03\x12\x18\n\x10update_timestamp\x18\t \x01(\x03\x12\x0f\n\x07percent\x18\n \x01(\x02\x12\x12\n\nlast_error\x18\x0b \x01(\t\"\xaa\x01\n\x12TaskMonitorStorage\x12=\n\x05tasks\x18\x01 \x03(\x0b\x32..ymir.controller.TaskMonitorStorage.TasksEntry\x1aU\n\nTasksEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\x36\n\x05value\x18\x02 \x01(\x0b\x32\'.ymir.controller.TaskMonitorStorageItem:\x02\x38\x01\"\xd9\x01\n\x16TaskMonitorStorageItem\x12\x33\n\x0cgeneral_info\x18\x01 \x01(\x0b\x32\x1d.ymir.controller.TaskInfoItem\x12\x19\n\x11monitor_file_path\x18\x02 \x01(\t\x12\x19\n\x11storage_file_path\x18\x03 \x01(\t\x12\x11\n\trepo_root\x18\x04 \x01(\t\x12\x13\n\x0b\x62ranch_name\x18\x05 \x01(\t\x12,\n\x07request\x18\x64 \x01(\x0b\x32\x1b.ymir.controller.GeneralReq*\xd0\x02\n\x0bRequestType\x12\x0b\n\x07UNKNOWN\x10\x00\x12\x12\n\x0e\x43MD_BRANCH_DEL\x10\x01\x12\x13\n\x0f\x43MD_BRANCH_LIST\x10\x02\x12\x17\n\x13\x43MD_BRANCH_CHECKOUT\x10\x03\x12\x15\n\x11\x43MD_BRANCH_CREATE\x10\x04\x12\r\n\tCMD_CLONE\x10\x05\x12\x0e\n\nCMD_COMMIT\x10\x06\x12\x0e\n\nCMD_FILTER\x10\x07\x12\x0c\n\x08\x43MD_INIT\x10\x08\x12\x0b\n\x07\x43MD_LOG\x10\t\x12\r\n\tCMD_MERGE\x10\n\x12\r\n\tUSER_LIST\x10\x65\x12\x0f\n\x0bUSER_CREATE\x10\x66\x12\x0f\n\x0bUSER_REMOVE\x10g\x12\r\n\tREPO_LIST\x10h\x12\x0f\n\x0bREPO_CREATE\x10i\x12\x0f\n\x0bREPO_REMOVE\x10j\x12\x10\n\x0bTASK_CREATE\x10\xe9\x07\x12\x0e\n\tTASK_INFO\x10\xea\x07\x32l\n\x16mir_controller_service\x12R\n\x13\x64\x61ta_manage_request\x12\x1b.ymir.controller.GeneralReq\x1a\x1c.ymir.controller.GeneralResp\"\x00\x62\x06proto3'
  ,
  dependencies=[mir__common__pb2.DESCRIPTOR,])

_REQUESTTYPE = _descriptor.EnumDescriptor(
  name='RequestType',
  full_name='ymir.controller.RequestType',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='UNKNOWN', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='CMD_BRANCH_DEL', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='CMD_BRANCH_LIST', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='CMD_BRANCH_CHECKOUT', index=3, number=3,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='CMD_BRANCH_CREATE', index=4, number=4,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='CMD_CLONE', index=5, number=5,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='CMD_COMMIT', index=6, number=6,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='CMD_FILTER', index=7, number=7,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='CMD_INIT', index=8, number=8,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='CMD_LOG', index=9, number=9,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='CMD_MERGE', index=10, number=10,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='USER_LIST', index=11, number=101,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='USER_CREATE', index=12, number=102,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='USER_REMOVE', index=13, number=103,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='REPO_LIST', index=14, number=104,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='REPO_CREATE', index=15, number=105,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='REPO_REMOVE', index=16, number=106,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='TASK_CREATE', index=17, number=1001,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='TASK_INFO', index=18, number=1002,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=2677,
  serialized_end=3013,
)
_sym_db.RegisterEnumDescriptor(_REQUESTTYPE)

RequestType = enum_type_wrapper.EnumTypeWrapper(_REQUESTTYPE)
UNKNOWN = 0
CMD_BRANCH_DEL = 1
CMD_BRANCH_LIST = 2
CMD_BRANCH_CHECKOUT = 3
CMD_BRANCH_CREATE = 4
CMD_CLONE = 5
CMD_COMMIT = 6
CMD_FILTER = 7
CMD_INIT = 8
CMD_LOG = 9
CMD_MERGE = 10
USER_LIST = 101
USER_CREATE = 102
USER_REMOVE = 103
REPO_LIST = 104
REPO_CREATE = 105
REPO_REMOVE = 106
TASK_CREATE = 1001
TASK_INFO = 1002



_GENERALREQ = _descriptor.Descriptor(
  name='GeneralReq',
  full_name='ymir.controller.GeneralReq',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='user_id', full_name='ymir.controller.GeneralReq.user_id', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='repo_id', full_name='ymir.controller.GeneralReq.repo_id', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='req_type', full_name='ymir.controller.GeneralReq.req_type', index=2,
      number=3, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='task_id', full_name='ymir.controller.GeneralReq.task_id', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='singleton_op', full_name='ymir.controller.GeneralReq.singleton_op', index=4,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='his_task_id', full_name='ymir.controller.GeneralReq.his_task_id', index=5,
      number=6, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='dst_task_id', full_name='ymir.controller.GeneralReq.dst_task_id', index=6,
      number=7, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='in_dataset_ids', full_name='ymir.controller.GeneralReq.in_dataset_ids', index=7,
      number=8, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='ex_dataset_ids', full_name='ymir.controller.GeneralReq.ex_dataset_ids', index=8,
      number=9, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='in_class_ids', full_name='ymir.controller.GeneralReq.in_class_ids', index=9,
      number=10, type=5, cpp_type=1, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='ex_class_ids', full_name='ymir.controller.GeneralReq.ex_class_ids', index=10,
      number=11, type=5, cpp_type=1, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='force', full_name='ymir.controller.GeneralReq.force', index=11,
      number=12, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='commit_message', full_name='ymir.controller.GeneralReq.commit_message', index=12,
      number=13, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='req_create_task', full_name='ymir.controller.GeneralReq.req_create_task', index=13,
      number=1001, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='req_get_task_info', full_name='ymir.controller.GeneralReq.req_get_task_info', index=14,
      number=1002, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=68,
  serialized_end=493,
)


_GENERALRESP = _descriptor.Descriptor(
  name='GeneralResp',
  full_name='ymir.controller.GeneralResp',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='code', full_name='ymir.controller.GeneralResp.code', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='req_task_id', full_name='ymir.controller.GeneralResp.req_task_id', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='message', full_name='ymir.controller.GeneralResp.message', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='ext_strs', full_name='ymir.controller.GeneralResp.ext_strs', index=3,
      number=4, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='resp_get_task_info', full_name='ymir.controller.GeneralResp.resp_get_task_info', index=4,
      number=1000, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=496,
  serialized_end=642,
)


_REQCREATETASK = _descriptor.Descriptor(
  name='ReqCreateTask',
  full_name='ymir.controller.ReqCreateTask',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='task_type', full_name='ymir.controller.ReqCreateTask.task_type', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='sampling_rate', full_name='ymir.controller.ReqCreateTask.sampling_rate', index=1,
      number=2, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='no_task_monitor', full_name='ymir.controller.ReqCreateTask.no_task_monitor', index=2,
      number=3, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='filter', full_name='ymir.controller.ReqCreateTask.filter', index=3,
      number=101, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='training', full_name='ymir.controller.ReqCreateTask.training', index=4,
      number=102, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='mining', full_name='ymir.controller.ReqCreateTask.mining', index=5,
      number=103, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='importing', full_name='ymir.controller.ReqCreateTask.importing', index=6,
      number=104, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='exporting', full_name='ymir.controller.ReqCreateTask.exporting', index=7,
      number=105, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='inference', full_name='ymir.controller.ReqCreateTask.inference', index=8,
      number=106, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='copy', full_name='ymir.controller.ReqCreateTask.copy', index=9,
      number=107, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=645,
  serialized_end=1108,
)


_TASKREQFILTER = _descriptor.Descriptor(
  name='TaskReqFilter',
  full_name='ymir.controller.TaskReqFilter',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='in_dataset_ids', full_name='ymir.controller.TaskReqFilter.in_dataset_ids', index=0,
      number=1, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='in_class_ids', full_name='ymir.controller.TaskReqFilter.in_class_ids', index=1,
      number=2, type=5, cpp_type=1, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='ex_class_ids', full_name='ymir.controller.TaskReqFilter.ex_class_ids', index=2,
      number=3, type=5, cpp_type=1, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1110,
  serialized_end=1193,
)


_TASKREQTRAINING_TRAININGDATASETTYPE = _descriptor.Descriptor(
  name='TrainingDatasetType',
  full_name='ymir.controller.TaskReqTraining.TrainingDatasetType',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='dataset_id', full_name='ymir.controller.TaskReqTraining.TrainingDatasetType.dataset_id', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='dataset_type', full_name='ymir.controller.TaskReqTraining.TrainingDatasetType.dataset_type', index=1,
      number=2, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1317,
  serialized_end=1402,
)

_TASKREQTRAINING = _descriptor.Descriptor(
  name='TaskReqTraining',
  full_name='ymir.controller.TaskReqTraining',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='in_dataset_types', full_name='ymir.controller.TaskReqTraining.in_dataset_types', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='in_class_ids', full_name='ymir.controller.TaskReqTraining.in_class_ids', index=1,
      number=2, type=5, cpp_type=1, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_TASKREQTRAINING_TRAININGDATASETTYPE, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1196,
  serialized_end=1402,
)


_TASKREQMINING = _descriptor.Descriptor(
  name='TaskReqMining',
  full_name='ymir.controller.TaskReqMining',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='in_dataset_ids', full_name='ymir.controller.TaskReqMining.in_dataset_ids', index=0,
      number=1, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='ex_dataset_ids', full_name='ymir.controller.TaskReqMining.ex_dataset_ids', index=1,
      number=2, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='model_hash', full_name='ymir.controller.TaskReqMining.model_hash', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='top_k', full_name='ymir.controller.TaskReqMining.top_k', index=3,
      number=4, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1404,
  serialized_end=1502,
)


_TASKREQIMPORTING = _descriptor.Descriptor(
  name='TaskReqImporting',
  full_name='ymir.controller.TaskReqImporting',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='asset_dir', full_name='ymir.controller.TaskReqImporting.asset_dir', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='annotation_dir', full_name='ymir.controller.TaskReqImporting.annotation_dir', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1504,
  serialized_end=1565,
)


_TASKREQEXPORTING = _descriptor.Descriptor(
  name='TaskReqExporting',
  full_name='ymir.controller.TaskReqExporting',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='dataset_id', full_name='ymir.controller.TaskReqExporting.dataset_id', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='format', full_name='ymir.controller.TaskReqExporting.format', index=1,
      number=2, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='asset_dir', full_name='ymir.controller.TaskReqExporting.asset_dir', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='annotation_dir', full_name='ymir.controller.TaskReqExporting.annotation_dir', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1567,
  serialized_end=1690,
)


_TASKREQINFERENCE = _descriptor.Descriptor(
  name='TaskReqInference',
  full_name='ymir.controller.TaskReqInference',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1692,
  serialized_end=1710,
)


_TASKREQCOPYDATA = _descriptor.Descriptor(
  name='TaskReqCopyData',
  full_name='ymir.controller.TaskReqCopyData',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='src_user_id', full_name='ymir.controller.TaskReqCopyData.src_user_id', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='src_repo_id', full_name='ymir.controller.TaskReqCopyData.src_repo_id', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='src_dataset_id', full_name='ymir.controller.TaskReqCopyData.src_dataset_id', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1712,
  serialized_end=1795,
)


_REQGETTASKINFO = _descriptor.Descriptor(
  name='ReqGetTaskInfo',
  full_name='ymir.controller.ReqGetTaskInfo',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='task_ids', full_name='ymir.controller.ReqGetTaskInfo.task_ids', index=0,
      number=1, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1797,
  serialized_end=1831,
)


_RESPGETTASKINFO_TASKINFOSENTRY = _descriptor.Descriptor(
  name='TaskInfosEntry',
  full_name='ymir.controller.RespGetTaskInfo.TaskInfosEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='ymir.controller.RespGetTaskInfo.TaskInfosEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='value', full_name='ymir.controller.RespGetTaskInfo.TaskInfosEntry.value', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=b'8\001',
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1922,
  serialized_end=2001,
)

_RESPGETTASKINFO = _descriptor.Descriptor(
  name='RespGetTaskInfo',
  full_name='ymir.controller.RespGetTaskInfo',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='task_infos', full_name='ymir.controller.RespGetTaskInfo.task_infos', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_RESPGETTASKINFO_TASKINFOSENTRY, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1834,
  serialized_end=2001,
)


_TASKINFOITEM = _descriptor.Descriptor(
  name='TaskInfoItem',
  full_name='ymir.controller.TaskInfoItem',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='task_id', full_name='ymir.controller.TaskInfoItem.task_id', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='base_task_id', full_name='ymir.controller.TaskInfoItem.base_task_id', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='user_id', full_name='ymir.controller.TaskInfoItem.user_id', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='repo_id', full_name='ymir.controller.TaskInfoItem.repo_id', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='type', full_name='ymir.controller.TaskInfoItem.type', index=4,
      number=5, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='state', full_name='ymir.controller.TaskInfoItem.state', index=5,
      number=6, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='start_timestamp', full_name='ymir.controller.TaskInfoItem.start_timestamp', index=6,
      number=7, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='finish_timestamp', full_name='ymir.controller.TaskInfoItem.finish_timestamp', index=7,
      number=8, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='update_timestamp', full_name='ymir.controller.TaskInfoItem.update_timestamp', index=8,
      number=9, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='percent', full_name='ymir.controller.TaskInfoItem.percent', index=9,
      number=10, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='last_error', full_name='ymir.controller.TaskInfoItem.last_error', index=10,
      number=11, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2004,
  serialized_end=2281,
)


_TASKMONITORSTORAGE_TASKSENTRY = _descriptor.Descriptor(
  name='TasksEntry',
  full_name='ymir.controller.TaskMonitorStorage.TasksEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='ymir.controller.TaskMonitorStorage.TasksEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='value', full_name='ymir.controller.TaskMonitorStorage.TasksEntry.value', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=b'8\001',
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2369,
  serialized_end=2454,
)

_TASKMONITORSTORAGE = _descriptor.Descriptor(
  name='TaskMonitorStorage',
  full_name='ymir.controller.TaskMonitorStorage',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='tasks', full_name='ymir.controller.TaskMonitorStorage.tasks', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_TASKMONITORSTORAGE_TASKSENTRY, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2284,
  serialized_end=2454,
)


_TASKMONITORSTORAGEITEM = _descriptor.Descriptor(
  name='TaskMonitorStorageItem',
  full_name='ymir.controller.TaskMonitorStorageItem',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='general_info', full_name='ymir.controller.TaskMonitorStorageItem.general_info', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='monitor_file_path', full_name='ymir.controller.TaskMonitorStorageItem.monitor_file_path', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='storage_file_path', full_name='ymir.controller.TaskMonitorStorageItem.storage_file_path', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='repo_root', full_name='ymir.controller.TaskMonitorStorageItem.repo_root', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='branch_name', full_name='ymir.controller.TaskMonitorStorageItem.branch_name', index=4,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='request', full_name='ymir.controller.TaskMonitorStorageItem.request', index=5,
      number=100, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2457,
  serialized_end=2674,
)

_GENERALREQ.fields_by_name['req_type'].enum_type = _REQUESTTYPE
_GENERALREQ.fields_by_name['req_create_task'].message_type = _REQCREATETASK
_GENERALREQ.fields_by_name['req_get_task_info'].message_type = _REQGETTASKINFO
_GENERALRESP.fields_by_name['resp_get_task_info'].message_type = _RESPGETTASKINFO
_REQCREATETASK.fields_by_name['task_type'].enum_type = mir__common__pb2._TASKTYPE
_REQCREATETASK.fields_by_name['filter'].message_type = _TASKREQFILTER
_REQCREATETASK.fields_by_name['training'].message_type = _TASKREQTRAINING
_REQCREATETASK.fields_by_name['mining'].message_type = _TASKREQMINING
_REQCREATETASK.fields_by_name['importing'].message_type = _TASKREQIMPORTING
_REQCREATETASK.fields_by_name['exporting'].message_type = _TASKREQEXPORTING
_REQCREATETASK.fields_by_name['inference'].message_type = _TASKREQINFERENCE
_REQCREATETASK.fields_by_name['copy'].message_type = _TASKREQCOPYDATA
_TASKREQTRAINING_TRAININGDATASETTYPE.fields_by_name['dataset_type'].enum_type = mir__common__pb2._TVTTYPE
_TASKREQTRAINING_TRAININGDATASETTYPE.containing_type = _TASKREQTRAINING
_TASKREQTRAINING.fields_by_name['in_dataset_types'].message_type = _TASKREQTRAINING_TRAININGDATASETTYPE
_TASKREQEXPORTING.fields_by_name['format'].enum_type = mir__common__pb2._LABELFORMAT
_RESPGETTASKINFO_TASKINFOSENTRY.fields_by_name['value'].message_type = _TASKINFOITEM
_RESPGETTASKINFO_TASKINFOSENTRY.containing_type = _RESPGETTASKINFO
_RESPGETTASKINFO.fields_by_name['task_infos'].message_type = _RESPGETTASKINFO_TASKINFOSENTRY
_TASKINFOITEM.fields_by_name['type'].enum_type = mir__common__pb2._TASKTYPE
_TASKINFOITEM.fields_by_name['state'].enum_type = mir__common__pb2._TASKSTATE
_TASKMONITORSTORAGE_TASKSENTRY.fields_by_name['value'].message_type = _TASKMONITORSTORAGEITEM
_TASKMONITORSTORAGE_TASKSENTRY.containing_type = _TASKMONITORSTORAGE
_TASKMONITORSTORAGE.fields_by_name['tasks'].message_type = _TASKMONITORSTORAGE_TASKSENTRY
_TASKMONITORSTORAGEITEM.fields_by_name['general_info'].message_type = _TASKINFOITEM
_TASKMONITORSTORAGEITEM.fields_by_name['request'].message_type = _GENERALREQ
DESCRIPTOR.message_types_by_name['GeneralReq'] = _GENERALREQ
DESCRIPTOR.message_types_by_name['GeneralResp'] = _GENERALRESP
DESCRIPTOR.message_types_by_name['ReqCreateTask'] = _REQCREATETASK
DESCRIPTOR.message_types_by_name['TaskReqFilter'] = _TASKREQFILTER
DESCRIPTOR.message_types_by_name['TaskReqTraining'] = _TASKREQTRAINING
DESCRIPTOR.message_types_by_name['TaskReqMining'] = _TASKREQMINING
DESCRIPTOR.message_types_by_name['TaskReqImporting'] = _TASKREQIMPORTING
DESCRIPTOR.message_types_by_name['TaskReqExporting'] = _TASKREQEXPORTING
DESCRIPTOR.message_types_by_name['TaskReqInference'] = _TASKREQINFERENCE
DESCRIPTOR.message_types_by_name['TaskReqCopyData'] = _TASKREQCOPYDATA
DESCRIPTOR.message_types_by_name['ReqGetTaskInfo'] = _REQGETTASKINFO
DESCRIPTOR.message_types_by_name['RespGetTaskInfo'] = _RESPGETTASKINFO
DESCRIPTOR.message_types_by_name['TaskInfoItem'] = _TASKINFOITEM
DESCRIPTOR.message_types_by_name['TaskMonitorStorage'] = _TASKMONITORSTORAGE
DESCRIPTOR.message_types_by_name['TaskMonitorStorageItem'] = _TASKMONITORSTORAGEITEM
DESCRIPTOR.enum_types_by_name['RequestType'] = _REQUESTTYPE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

GeneralReq = _reflection.GeneratedProtocolMessageType('GeneralReq', (_message.Message,), {
  'DESCRIPTOR' : _GENERALREQ,
  '__module__' : 'mir_controller_service_pb2'
  # @@protoc_insertion_point(class_scope:ymir.controller.GeneralReq)
  })
_sym_db.RegisterMessage(GeneralReq)

GeneralResp = _reflection.GeneratedProtocolMessageType('GeneralResp', (_message.Message,), {
  'DESCRIPTOR' : _GENERALRESP,
  '__module__' : 'mir_controller_service_pb2'
  # @@protoc_insertion_point(class_scope:ymir.controller.GeneralResp)
  })
_sym_db.RegisterMessage(GeneralResp)

ReqCreateTask = _reflection.GeneratedProtocolMessageType('ReqCreateTask', (_message.Message,), {
  'DESCRIPTOR' : _REQCREATETASK,
  '__module__' : 'mir_controller_service_pb2'
  # @@protoc_insertion_point(class_scope:ymir.controller.ReqCreateTask)
  })
_sym_db.RegisterMessage(ReqCreateTask)

TaskReqFilter = _reflection.GeneratedProtocolMessageType('TaskReqFilter', (_message.Message,), {
  'DESCRIPTOR' : _TASKREQFILTER,
  '__module__' : 'mir_controller_service_pb2'
  # @@protoc_insertion_point(class_scope:ymir.controller.TaskReqFilter)
  })
_sym_db.RegisterMessage(TaskReqFilter)

TaskReqTraining = _reflection.GeneratedProtocolMessageType('TaskReqTraining', (_message.Message,), {

  'TrainingDatasetType' : _reflection.GeneratedProtocolMessageType('TrainingDatasetType', (_message.Message,), {
    'DESCRIPTOR' : _TASKREQTRAINING_TRAININGDATASETTYPE,
    '__module__' : 'mir_controller_service_pb2'
    # @@protoc_insertion_point(class_scope:ymir.controller.TaskReqTraining.TrainingDatasetType)
    })
  ,
  'DESCRIPTOR' : _TASKREQTRAINING,
  '__module__' : 'mir_controller_service_pb2'
  # @@protoc_insertion_point(class_scope:ymir.controller.TaskReqTraining)
  })
_sym_db.RegisterMessage(TaskReqTraining)
_sym_db.RegisterMessage(TaskReqTraining.TrainingDatasetType)

TaskReqMining = _reflection.GeneratedProtocolMessageType('TaskReqMining', (_message.Message,), {
  'DESCRIPTOR' : _TASKREQMINING,
  '__module__' : 'mir_controller_service_pb2'
  # @@protoc_insertion_point(class_scope:ymir.controller.TaskReqMining)
  })
_sym_db.RegisterMessage(TaskReqMining)

TaskReqImporting = _reflection.GeneratedProtocolMessageType('TaskReqImporting', (_message.Message,), {
  'DESCRIPTOR' : _TASKREQIMPORTING,
  '__module__' : 'mir_controller_service_pb2'
  # @@protoc_insertion_point(class_scope:ymir.controller.TaskReqImporting)
  })
_sym_db.RegisterMessage(TaskReqImporting)

TaskReqExporting = _reflection.GeneratedProtocolMessageType('TaskReqExporting', (_message.Message,), {
  'DESCRIPTOR' : _TASKREQEXPORTING,
  '__module__' : 'mir_controller_service_pb2'
  # @@protoc_insertion_point(class_scope:ymir.controller.TaskReqExporting)
  })
_sym_db.RegisterMessage(TaskReqExporting)

TaskReqInference = _reflection.GeneratedProtocolMessageType('TaskReqInference', (_message.Message,), {
  'DESCRIPTOR' : _TASKREQINFERENCE,
  '__module__' : 'mir_controller_service_pb2'
  # @@protoc_insertion_point(class_scope:ymir.controller.TaskReqInference)
  })
_sym_db.RegisterMessage(TaskReqInference)

TaskReqCopyData = _reflection.GeneratedProtocolMessageType('TaskReqCopyData', (_message.Message,), {
  'DESCRIPTOR' : _TASKREQCOPYDATA,
  '__module__' : 'mir_controller_service_pb2'
  # @@protoc_insertion_point(class_scope:ymir.controller.TaskReqCopyData)
  })
_sym_db.RegisterMessage(TaskReqCopyData)

ReqGetTaskInfo = _reflection.GeneratedProtocolMessageType('ReqGetTaskInfo', (_message.Message,), {
  'DESCRIPTOR' : _REQGETTASKINFO,
  '__module__' : 'mir_controller_service_pb2'
  # @@protoc_insertion_point(class_scope:ymir.controller.ReqGetTaskInfo)
  })
_sym_db.RegisterMessage(ReqGetTaskInfo)

RespGetTaskInfo = _reflection.GeneratedProtocolMessageType('RespGetTaskInfo', (_message.Message,), {

  'TaskInfosEntry' : _reflection.GeneratedProtocolMessageType('TaskInfosEntry', (_message.Message,), {
    'DESCRIPTOR' : _RESPGETTASKINFO_TASKINFOSENTRY,
    '__module__' : 'mir_controller_service_pb2'
    # @@protoc_insertion_point(class_scope:ymir.controller.RespGetTaskInfo.TaskInfosEntry)
    })
  ,
  'DESCRIPTOR' : _RESPGETTASKINFO,
  '__module__' : 'mir_controller_service_pb2'
  # @@protoc_insertion_point(class_scope:ymir.controller.RespGetTaskInfo)
  })
_sym_db.RegisterMessage(RespGetTaskInfo)
_sym_db.RegisterMessage(RespGetTaskInfo.TaskInfosEntry)

TaskInfoItem = _reflection.GeneratedProtocolMessageType('TaskInfoItem', (_message.Message,), {
  'DESCRIPTOR' : _TASKINFOITEM,
  '__module__' : 'mir_controller_service_pb2'
  # @@protoc_insertion_point(class_scope:ymir.controller.TaskInfoItem)
  })
_sym_db.RegisterMessage(TaskInfoItem)

TaskMonitorStorage = _reflection.GeneratedProtocolMessageType('TaskMonitorStorage', (_message.Message,), {

  'TasksEntry' : _reflection.GeneratedProtocolMessageType('TasksEntry', (_message.Message,), {
    'DESCRIPTOR' : _TASKMONITORSTORAGE_TASKSENTRY,
    '__module__' : 'mir_controller_service_pb2'
    # @@protoc_insertion_point(class_scope:ymir.controller.TaskMonitorStorage.TasksEntry)
    })
  ,
  'DESCRIPTOR' : _TASKMONITORSTORAGE,
  '__module__' : 'mir_controller_service_pb2'
  # @@protoc_insertion_point(class_scope:ymir.controller.TaskMonitorStorage)
  })
_sym_db.RegisterMessage(TaskMonitorStorage)
_sym_db.RegisterMessage(TaskMonitorStorage.TasksEntry)

TaskMonitorStorageItem = _reflection.GeneratedProtocolMessageType('TaskMonitorStorageItem', (_message.Message,), {
  'DESCRIPTOR' : _TASKMONITORSTORAGEITEM,
  '__module__' : 'mir_controller_service_pb2'
  # @@protoc_insertion_point(class_scope:ymir.controller.TaskMonitorStorageItem)
  })
_sym_db.RegisterMessage(TaskMonitorStorageItem)


_RESPGETTASKINFO_TASKINFOSENTRY._options = None
_TASKMONITORSTORAGE_TASKSENTRY._options = None

_MIR_CONTROLLER_SERVICE = _descriptor.ServiceDescriptor(
  name='mir_controller_service',
  full_name='ymir.controller.mir_controller_service',
  file=DESCRIPTOR,
  index=0,
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_start=3015,
  serialized_end=3123,
  methods=[
  _descriptor.MethodDescriptor(
    name='data_manage_request',
    full_name='ymir.controller.mir_controller_service.data_manage_request',
    index=0,
    containing_service=None,
    input_type=_GENERALREQ,
    output_type=_GENERALRESP,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
])
_sym_db.RegisterServiceDescriptor(_MIR_CONTROLLER_SERVICE)

DESCRIPTOR.services_by_name['mir_controller_service'] = _MIR_CONTROLLER_SERVICE

# @@protoc_insertion_point(module_scope)



import logging

from mir import scm
from mir.commands import base
from mir.tools import checker, code


class CmdCheckout(base.BaseCommand):
    @staticmethod
    def run_with_args(mir_root: str, commit_id: str, branch_new: bool = False) -> int:
        return_code = checker.check(mir_root)
        if return_code != code.MirCode.RC_OK:
            return return_code

        repo_git = scm.Scm(root_dir=mir_root, scm_executable='git')
        repo_dvc = scm.Scm(root_dir=mir_root, scm_executable='dvc')

        # git options
        cmd_opts = []
        if branch_new:
            cmd_opts.append("-b")
        cmd_opts.append(commit_id)

        output_str = repo_git.checkout(cmd_opts)
        if output_str:
            logging.info("\n%s" % output_str)

        output_str = repo_dvc.checkout("-f")
        if output_str:
            logging.info("\n%s" % output_str)

        return code.MirCode.RC_OK

    def run(self):
        logging.debug("command checkout: %s", self.args)

        return CmdCheckout.run_with_args(mir_root=self.args.mir_root,
                                         branch_new=self.args.branch_new,
                                         commit_id=self.args.commit_id)


def bind_to_subparsers(subparsers, parent_parser):
    checkout_arg_parser = subparsers.add_parser("checkout",
                                                aliases=["co"],
                                                parents=[parent_parser],
                                                description="use this command to show mir repo branches",
                                                help="checkout commits or add new branches")
    checkout_arg_parser.add_argument("-b", dest="branch_new", help="add a new branch", action="store_true")
    checkout_arg_parser.add_argument("commit_id", help="destination commit id or tag name", nargs=1, type=str)
    checkout_arg_parser.set_defaults(func=CmdCheckout)



"""
mir merge: merge contents from another guest branch to current branch
"""

import datetime
import logging
from typing import Dict, Tuple

from mir.commands import base
from mir.tools import checker, code, mir_storage, mir_storage_ops, revs_parser
from ymir_proto.gen import mir_common_pb2 as mir_common
from ymir_proto.gen import mir_entities_pb2 as mirpb


def _match_asset_ids(host_ids: set, guest_ids: set) -> Tuple[set, set, set]:
    """
    match asset ids

    Args:
        host_ids (set): host ids
        guest_ids (set): guest ids

    Returns:
        Tuple[set, set, set]: host_only_ids, guest_only_ids, joint_ids
    """
    insets = host_ids & guest_ids
    return (host_ids - insets, guest_ids - insets, insets)


def _merge_metadatas(host_mir_metadatas: mirpb.MirMetadatas, guest_mir_metadatas: mirpb.MirMetadatas,
                     id_guest_only: set, id_joint: set, suggested_tvt_type: mir_common.TvtType, strategy: str):
    if not host_mir_metadatas or not guest_mir_metadatas:
        raise ValueError("input host/guest mir_metadatas is invalid")

    # for all ids only in `guest_mir_metadatas`, add them to `host_mir_metadatas`
    for asset_id in id_guest_only:
        host_mir_metadatas.attributes[asset_id].CopyFrom(guest_mir_metadatas.attributes[asset_id])
        if suggested_tvt_type != mir_common.TvtTypeUnknown:
            host_mir_metadatas.attributes[asset_id].tvt_type = suggested_tvt_type
    # for all ids in both two mir_metadatas
    if strategy == "guest":
        for asset_id in id_joint:
            host_mir_metadatas.attributes[asset_id].CopyFrom(guest_mir_metadatas.attributes[asset_id])
            if suggested_tvt_type != mir_common.TvtTypeUnknown:
                host_mir_metadatas.attributes[asset_id].tvt_type = suggested_tvt_type
    elif strategy == "host":
        pass
    elif strategy == "stop":
        if len(id_joint) > 0:
            raise RuntimeError("joint ids detected, stop")
    else:
        raise ValueError("invalid strategy")


# WARNING: annotations are filtered in task level, NOT in image level.
def _merge_annotations(host_mir_annotations: mirpb.MirAnnotations, guest_mir_annotations: mirpb.MirAnnotations,
                       strategy: str):
    """
    add all annotations in guest_mir_annotations into host_mir_annotations

    Args:
        host_mir_annotations (mirpb.MirAnnotations): host annotations
        guest_mir_annotations (mirpb.MirAnnotations): guest annotations
        strategy (str): host, guest, stop

    Raises:
        ValueError: if host or guest annotations empty
        ValueError: if conflicts occured in strategy stop
    """
    if not host_mir_annotations or not guest_mir_annotations:
        raise ValueError("input host/guest mir_annotations is invalid")
    if not host_mir_annotations.head_task_id:
        raise ValueError("no head_task_id found in host_mir_annotations")
    if len(guest_mir_annotations.task_annotations) == 0:
        logging.warning('empty guest_mir_annotations')
        return

    task_id = host_mir_annotations.head_task_id
    host_image_annotations = host_mir_annotations.task_annotations[
        host_mir_annotations.head_task_id].image_annotations  # type: Dict[str, mirpb.SingleImageAnnotations]
    guest_image_annotations = guest_mir_annotations.task_annotations[
        guest_mir_annotations.head_task_id].image_annotations  # type: Dict[str, mirpb.SingleImageAnnotations]
    host_only_ids, guest_only_ids, joint_ids = _match_asset_ids(set(host_image_annotations.keys()),
                                                                set(guest_image_annotations.keys()))

    if strategy == "stop" and joint_ids:
        raise ValueError("found conflicts; abort")

    for asset_id in host_only_ids:
        host_mir_annotations.task_annotations[task_id].image_annotations[asset_id].CopyFrom(
            host_image_annotations[asset_id])
    for asset_id in guest_only_ids:
        host_mir_annotations.task_annotations[task_id].image_annotations[asset_id].CopyFrom(
            guest_image_annotations[asset_id])
    for asset_id in joint_ids:
        if strategy.lower() == "host":
            if asset_id not in host_mir_annotations.task_annotations[task_id].image_annotations:
                host_mir_annotations.task_annotations[task_id].image_annotations[asset_id].CopyFrom(
                    host_image_annotations[asset_id])
        elif strategy.lower() == "guest":
            host_mir_annotations.task_annotations[task_id].image_annotations[asset_id].CopyFrom(
                guest_image_annotations[asset_id])


def _get_union_keywords(host_keywords, guest_keywords, strategy):
    if host_keywords:
        host_keywords_set = set(host_keywords)
    else:
        host_keywords_set = set()
    if guest_keywords:
        guest_keywords_set = set(guest_keywords)
    else:
        guest_keywords_set = set()

    if strategy == "host":
        merged_keywords_set = host_keywords_set
    elif strategy == "guest":
        merged_keywords_set = guest_keywords_set
    else:
        raise Exception("unknown strategy type: " + strategy)
    return merged_keywords_set


def _merge_keywords(host_mir_keywords, guest_mir_keywords, id_guest_only: set, id_joint: set, strategy: str):
    if host_mir_keywords is None or guest_mir_keywords is None:
        raise RuntimeError('Invalid keywords message map.')

    for asset_id in id_guest_only:
        if asset_id not in guest_mir_keywords:
            continue
        host_mir_keywords[asset_id].predifined_keyids[:] = guest_mir_keywords[asset_id].predifined_keyids
        host_mir_keywords[asset_id].customized_keywords[:] = guest_mir_keywords[asset_id].customized_keywords

    for asset_id in id_joint:
        if asset_id not in guest_mir_keywords:
            continue

        merged_keywords_set_predifined = _get_union_keywords(host_mir_keywords[asset_id].predifined_keyids,
                                                             guest_mir_keywords[asset_id].predifined_keyids, strategy)

        if merged_keywords_set_predifined:
            host_mir_keywords[asset_id].predifined_keyids[:] = merged_keywords_set_predifined

        merged_keywords_set_customized = _get_union_keywords(host_mir_keywords[asset_id].customized_keywords,
                                                             guest_mir_keywords[asset_id].customized_keywords, strategy)

        if merged_keywords_set_customized:
            host_mir_keywords[asset_id].customized_keywords[:] = merged_keywords_set_customized


def _merge_tasks(host_mir_tasks, guest_mir_tasks):
    if not host_mir_tasks or not guest_mir_tasks:
        raise ValueError("input host/guest mir_tasks is invalid")

    for guest_task_id, guest_task in guest_mir_tasks.tasks.items():
        if guest_task_id not in host_mir_tasks.tasks:
            host_mir_tasks.tasks[guest_task_id].CopyFrom(guest_task)


def _tvt_type_from_str(typ: str) -> int:
    if typ == "tr":
        return mir_common.TvtTypeTraining
    elif typ == "va":
        return mir_common.TvtTypeValidation
    elif typ == "te":
        return mir_common.TvtTypeTest
    elif not typ:
        return mir_common.TvtTypeUnknown
    else:
        raise ValueError(f"invalid typ: {typ}")


def _merge_to_mir(host_mir_metadatas: mirpb.MirMetadatas, host_mir_annotations: mirpb.MirAnnotations,
                  host_mir_keywords: mirpb.MirKeywords, host_mir_tasks: mirpb.MirTasks, mir_root: str,
                  guest_typ_rev_tid: revs_parser.TypRevTid, strategy: str) -> int:
    """
    merge contents in `guest_typ_rev_tid` to `host_mir_xxx`

    Args:
        host_mir_metadatas (mirpb.MirMetadatas): host metadatas
        host_mir_annotations (mirpb.MirAnnotations): host annotations
        host_mir_keywords (mirpb.MirKeywords): host keywords
        host_mir_tasks (mirpb.MirTasks): host tasks
        mir_root (str): path to mir repo
        guest_typ_rev_tid (revs_parser.TypRevTid): guest typ:rev@tid
        strategy (str): host / guest / stop

    Raises:
        RuntimeError: when guest branch has no metadatas, or guest branch has no tasks
    """
    mir_data = mir_storage_ops.MirStorageOps.load(mir_root=mir_root,
                                                  mir_branch=guest_typ_rev_tid.rev,
                                                  mir_storages=mir_storage.get_all_mir_storage())

    guest_mir_metadatas = mir_data.get(mir_common.MirStorage.MIR_METADATAS, None)
    if not guest_mir_metadatas:
        raise RuntimeError(f"guest repo {mir_root}:{guest_typ_rev_tid.rev} has no metadata.")

    id_host_only, id_guest_only, id_joint = _match_asset_ids(set(host_mir_metadatas.attributes.keys()),
                                                             set(guest_mir_metadatas.attributes.keys()))

    logging.info(f"{guest_typ_rev_tid} host assets: {len(id_host_only)}, guest assets: {len(id_guest_only)}, "
                 f"joint assets: {len(id_joint)}")

    # if no guest_only assets found, return
    if not id_guest_only and not id_joint:
        return mir_common.RC_OK
    if strategy.lower() == "stop" and id_joint:
        logging.warning("found conflict on merge strategy STOP: abort")
        return code.MirCode.RC_ERROR_CONFLICTS_OCCURED

    _merge_metadatas(host_mir_metadatas=host_mir_metadatas,
                     guest_mir_metadatas=guest_mir_metadatas,
                     id_guest_only=id_guest_only,
                     id_joint=id_joint,
                     suggested_tvt_type=_tvt_type_from_str(guest_typ_rev_tid.typ),
                     strategy=strategy)

    guest_mir_annotations = mir_data.get(mir_common.MirStorage.MIR_ANNOTATIONS, None)
    if not guest_mir_annotations:
        logging.warning("guest repo {}:{} has no annotations.".format(mir_root, guest_typ_rev_tid.rev))
    _merge_annotations(host_mir_annotations=host_mir_annotations,
                       guest_mir_annotations=guest_mir_annotations,
                       strategy=strategy)

    guest_mir_keywords = mir_data.get(mir_common.MirStorage.MIR_KEYWORDS, None)
    if not guest_mir_keywords:
        logging.warning(f"guest repo {mir_root}:{guest_typ_rev_tid.rev} has no keywords.")
    _merge_keywords(host_mir_keywords.keywords, guest_mir_keywords.keywords, id_guest_only, id_joint, strategy)

    guest_mir_tasks = mir_data.get(mir_common.MirStorage.MIR_TASKS, None)
    if not guest_mir_tasks:
        raise RuntimeError(f"guest repo {mir_root}:{guest_typ_rev_tid.rev} has no tasks.")
    _merge_tasks(host_mir_tasks, guest_mir_tasks)

    return code.MirCode.RC_OK


def _exclude_from_mir(host_mir_metadatas: mirpb.MirMetadatas, host_mir_annotations: mirpb.MirAnnotations,
                      host_mir_keywords: mirpb.MirKeywords, mir_root: str, branch_id: str):
    if not branch_id:
        raise RuntimeError('empty exclude id')
    if not mir_root:
        mir_root = '.'
    if not host_mir_metadatas:
        raise RuntimeError('invalid host_mir_metadatas')

    guest_mir_metadatas = mir_storage_ops.MirStorageOps.load_single(mir_root=mir_root,
                                                                    mir_branch=branch_id,
                                                                    ms=mir_common.MirStorage.MIR_METADATAS)
    if not guest_mir_metadatas:
        raise RuntimeError("guest repo {}:{} has no metadata.".format(mir_root, branch_id))

    _, _, id_joint = _match_asset_ids(set(host_mir_metadatas.attributes.keys()),
                                      set(guest_mir_metadatas.attributes.keys()))
    for asset_id in id_joint:
        del host_mir_metadatas.attributes[asset_id]

        if asset_id in host_mir_annotations.task_annotations[host_mir_annotations.head_task_id].image_annotations:
            del host_mir_annotations.task_annotations[host_mir_annotations.head_task_id].image_annotations[asset_id]
        if asset_id in host_mir_keywords.keywords:
            del host_mir_keywords.keywords[asset_id]
    return code.MirCode.RC_OK


class CmdMerge(base.BaseCommand):
    def run(self):
        logging.debug("command merge: %s", self.args)
        return CmdMerge.run_with_args(mir_root=self.args.mir_root,
                                      src_revs=self.args.src_revs,
                                      ex_src_revs=self.args.ex_src_revs,
                                      dst_rev=self.args.dst_rev,
                                      strategy=self.args.strategy)

    @staticmethod
    def run_with_args(mir_root: str, src_revs: str, ex_src_revs: str, dst_rev: str, strategy: str):
        if not mir_root:
            logging.error("{} not in the mir repo!".format(mir_root))
            return code.MirCode.RC_ERROR_INVALID_MIR_REPO

        if not src_revs or not dst_rev:
            logging.error("empty --src-revs or --dst-rev")
            return code.MirCode.RC_ERROR_INVALID_ARGS
        src_typ_rev_tids = revs_parser.parse_arg_revs(src_revs)
        if not src_typ_rev_tids:
            logging.error(f"no source provided in --src-revs: {src_revs}")
            return code.MirCode.RC_ERROR_INVALID_ARGS
        dst_typ_rev_tid = revs_parser.parse_single_arg_rev(dst_rev)
        if not dst_typ_rev_tid.rev or not dst_typ_rev_tid.tid:
            logging.error(f"invalid --dst-rev: {dst_rev}, needs rev@tid")
            return code.MirCode.RC_ERROR_INVALID_ARGS

        return_code = checker.check(mir_root)
        if return_code != code.MirCode.RC_OK:
            return return_code

        # Read host id mir data.
        host_mir_metadatas = mirpb.MirMetadatas()
        host_mir_annotations = mirpb.MirAnnotations()
        host_mir_keywords = mirpb.MirKeywords()
        host_mir_tasks = mirpb.MirTasks()

        host_mir_annotations.head_task_id = dst_typ_rev_tid.tid

        for typ_rev_tid in src_typ_rev_tids:
            ret = _merge_to_mir(host_mir_metadatas=host_mir_metadatas,
                                host_mir_annotations=host_mir_annotations,
                                host_mir_keywords=host_mir_keywords,
                                host_mir_tasks=host_mir_tasks,
                                mir_root=mir_root,
                                guest_typ_rev_tid=typ_rev_tid,
                                strategy=strategy)
            if ret != code.MirCode.RC_OK:
                return ret

        ex_typ_rev_tids = revs_parser.parse_arg_revs(ex_src_revs)
        for typ_rev_tid in ex_typ_rev_tids:
            ret = _exclude_from_mir(host_mir_metadatas=host_mir_metadatas,
                                    host_mir_annotations=host_mir_annotations,
                                    host_mir_keywords=host_mir_keywords,
                                    mir_root=mir_root,
                                    branch_id=typ_rev_tid.rev)
            if ret != code.MirCode.RC_OK:
                return ret

        # create and write tasks
        task = mirpb.Task()
        task.type = mir_common.TaskTypeMerge
        task.name = f"merge: {src_revs} - {ex_src_revs} to {dst_rev}"
        task.task_id = dst_typ_rev_tid.tid
        task.timestamp = int(datetime.datetime.now().timestamp())
        mir_storage_ops.add_mir_task(host_mir_tasks, task)

        host_typ_rev_tid = src_typ_rev_tids[0]
        mir_data = {
            mir_common.MirStorage.MIR_METADATAS: host_mir_metadatas,
            mir_common.MirStorage.MIR_ANNOTATIONS: host_mir_annotations,
            mir_common.MirStorage.MIR_KEYWORDS: host_mir_keywords,
            mir_common.MirStorage.MIR_TASKS: host_mir_tasks,
        }
        mir_storage_ops.MirStorageOps.save_and_commit(mir_root=mir_root,
                                                      mir_branch=dst_typ_rev_tid.rev,
                                                      his_branch=host_typ_rev_tid.rev,
                                                      mir_datas=mir_data,
                                                      commit_message=task.name)

        logging.debug("mir merge: write files done")

        return code.MirCode.RC_OK


def bind_to_subparsers(subparsers, parent_parser):  # pragma: no cover
    merge_arg_parser = subparsers.add_parser("merge",
                                             parents=[parent_parser],
                                             description="use this command to merge contents from other branch",
                                             help="merge contents from other branch")
    merge_arg_parser.add_argument("--src-revs",
                                  dest="src_revs",
                                  type=str,
                                  required=True,
                                  help="source tvt types, revs and base task ids, first the host, others the guests, "
                                  "can begin with tr:/va:/te:, uses own tvt type if no prefix assigned")
    merge_arg_parser.add_argument("--ex-src-revs",
                                  dest="ex_src_revs",
                                  type=str,
                                  help="branch(es) id, from which you want to exclude, seperated by comma.")
    merge_arg_parser.add_argument("--dst-rev",
                                  dest="dst_rev",
                                  type=str,
                                  required=True,
                                  help="rev@tid: destination branch name and task id")
    merge_arg_parser.add_argument("-s",
                                  dest="strategy",
                                  type=str,
                                  default="stop",
                                  choices=["stop", "host", "guest"],
                                  help="conflict resolvation strategy, stop (default): stop when conflict detects; "
                                  "host: use host; guest: use guest")
    merge_arg_parser.set_defaults(func=CmdMerge)



import datetime
import logging
import os
import shutil

from mir.commands import base
from mir.tools import annotations, code, hash_utils, metadatas, mir_storage_ops, revs_parser
from ymir_proto.gen import mir_common_pb2 as mir_common
from ymir_proto.gen import mir_entities_pb2 as mirpb


class CmdImport(base.BaseCommand):
    def run(self):
        logging.debug("command import: %s", self.args)

        return CmdImport.run_with_args(mir_root=self.args.mir_root,
                                       index_abs=self.args.index_file,
                                       anno_abs=self.args.anno,
                                       gen_abs=self.args.gen,
                                       dataset_name=self.args.dataset_name,
                                       dst_rev=self.args.dst_rev,
                                       src_revs=self.args.src_revs)

    @staticmethod
    def run_with_args(mir_root: str, index_abs: str, anno_abs: str, gen_abs: str, dataset_name: str, dst_rev: str,
                      src_revs: str):
        # Step 1: check args and prepare environment.
        if not index_abs or not anno_abs or not gen_abs:
            logging.error("Missing input args")
            return code.MirCode.RC_ERROR_INVALID_ARGS
        if not os.path.isdir(anno_abs) or not os.path.isfile(index_abs):
            logging.error("Incorrect input paths")
            return code.MirCode.RC_ERROR_INVALID_ARGS
        if not dst_rev:
            logging.error("empty --dst-rev")
            return code.MirCode.RC_ERROR_INVALID_ARGS
        dst_typ_rev_tid = revs_parser.parse_single_arg_rev(dst_rev)
        if not dst_typ_rev_tid.rev or not dst_typ_rev_tid.tid:
            logging.error(f"destination branch or task id not provided: {dst_rev}")
            return code.MirCode.RC_ERROR_INVALID_ARGS
        if not dataset_name:
            dataset_name = dst_typ_rev_tid.tid

        # Step 2: generate sha1 file and rename images.
        # sha1 file to be written.
        sha1_index_abs = os.path.join(gen_abs, os.path.basename(index_abs) + '.sha1')
        sha1_folder_abs = os.path.join(gen_abs, 'hashed')
        ret = _generate_sha_and_cp(index_abs, sha1_index_abs, sha1_folder_abs)
        if ret != code.MirCode.RC_OK:
            logging.error("importing sha1 error.")
            return ret

        # Step 3 import metadat and annotations:
        mir_metadatas = mirpb.MirMetadatas()
        ret = metadatas.import_metadatas(mir_metadatas=mir_metadatas,
                                         dataset_name=dataset_name,
                                         in_sha1_path=sha1_index_abs,
                                         hashed_asset_root=sha1_folder_abs)
        if ret != code.MirCode.RC_OK:
            logging.error("importing metadatas error.")
            return ret

        mir_annotation = mirpb.MirAnnotations()
        mir_keywords = mirpb.MirKeywords()
        ret = annotations.import_annotations(mir_annotation=mir_annotation,
                                             mir_keywords=mir_keywords,
                                             in_sha1_file=sha1_index_abs,
                                             annotations_dir_path=anno_abs,
                                             task_id=dst_typ_rev_tid.tid)
        if ret != code.MirCode.RC_OK:
            logging.error("importing annotations error.")
            return ret

        # create and write tasks
        mir_tasks = mirpb.MirTasks()
        task = mirpb.Task()
        task.type = mir_common.TaskTypeImportData
        task.name = f"importing {index_abs}-{anno_abs}-{gen_abs} as {dataset_name}"
        task.task_id = dst_typ_rev_tid.tid
        task.timestamp = int(datetime.datetime.now().timestamp())
        mir_storage_ops.add_mir_task(mir_tasks, task)

        mir_data = {
            mir_common.MirStorage.MIR_METADATAS: mir_metadatas,
            mir_common.MirStorage.MIR_ANNOTATIONS: mir_annotation,
            mir_common.MirStorage.MIR_KEYWORDS: mir_keywords,
            mir_common.MirStorage.MIR_TASKS: mir_tasks,
        }
        mir_storage_ops.MirStorageOps.save_and_commit(mir_root=mir_root,
                                                      his_branch=(src_revs or "master"),
                                                      mir_branch=dst_typ_rev_tid.rev,
                                                      mir_datas=mir_data,
                                                      commit_message=dst_typ_rev_tid.tid)

        logging.info("import done.")

        return code.MirCode.RC_OK


def _generate_sha_and_cp(index_file, sha_idx_file, sha_folder):
    if not os.path.isfile(index_file):
        logging.error("Invalid input paths")
        return code.MirCode.RC_ERROR_INVALID_ARGS

    os.makedirs(sha_folder, exist_ok=True)

    with open(index_file) as idx_f, open(sha_idx_file, 'w') as sha_f:
        idx = 0
        for line in idx_f:
            media_src = line.strip()
            if not media_src or not os.path.isfile(media_src):
                logging.warning("invalid file: ", media_src)
                continue
            sha1 = hash_utils.sha1sum_for_file(media_src)
            sha_f.writelines("\t".join([sha1, media_src]) + '\n')

            media_dst = os.path.join(sha_folder, sha1)
            if not os.path.isfile(media_dst):
                shutil.copyfile(media_src, media_dst)

            idx += 1
            if idx % 5000 == 0:
                logging.info("finished %d hashes", idx)
    return code.MirCode.RC_OK


def bind_to_subparsers(subparsers, parent_parser):  # pragma: no cover
    importing_arg_parser = subparsers.add_parser("import",
                                                 parents=[parent_parser],
                                                 description="use this command to import data from img/anno folder",
                                                 help="import raw data")
    importing_arg_parser.add_argument("--index-file",
                                      dest="index_file",
                                      type=str,
                                      help="index of input media, one file per line")
    importing_arg_parser.add_argument("--annotation-dir", dest="anno", type=str, help="corresponding annotation folder")
    importing_arg_parser.add_argument("--gen-dir", dest="gen", type=str, help="storage path of generated data files")
    importing_arg_parser.add_argument("--dataset-name",
                                      dest="dataset_name",
                                      type=str,
                                      help="name of the dataset to be created, use tid if not set.")
    importing_arg_parser.add_argument("--src-revs", dest="src_revs", type=str, help="rev: source rev")
    importing_arg_parser.add_argument("--dst-rev",
                                      dest="dst_rev",
                                      type=str,
                                      required=True,
                                      help="rev@tid: destination branch name and task id")
    importing_arg_parser.set_defaults(func=CmdImport)



import logging
import os
from typing import Tuple

import xml.dom.minidom

from mir.tools import code
from ymir_proto.gen import mir_entities_pb2 as mirpb
from ymir_proto.util import ids


def _get_dom_xml_tag_node(node: xml.dom.minidom.Element, tag_name: str) -> xml.dom.minidom.Element:
    """
    suppose we have the following xml:
    ```
    <blabla>
        <tag1>tag1_value</tag1>
        <tag2>tag2_value</tag2>
    </blabla>
    ```
    and we have node point to <blabla>, we can use this function to get node tag1 and tag2 \n
    if tag not found, returns None
    """
    tag_nodes = node.getElementsByTagName(tag_name)
    if len(tag_nodes) > 0 and len(tag_nodes[0].childNodes) > 0:
        return tag_nodes[0]
    return None


def _get_dom_xml_tag_data(node: xml.dom.minidom.Element, tag_name: str) -> str:
    """
    suppose we have the following xml:
    ```
    <blabla>
        <tag1>tag1_value</tag1>
        <tag2>tag2_value</tag2>
    </blabla>
    ```
    and we have node point to <blabla>, we can use this function to get tag1_value and tag2_value \n
    if tag not found, returns None
    """
    tag_node = _get_dom_xml_tag_node(node, tag_name)
    if tag_node and len(tag_node.childNodes) > 0:
        return tag_node.childNodes[0].data
    return None


def _xml_obj_to_annotation_and_name(obj: xml.dom.minidom.Element,
                                    class_type_manager: ids.ClassIdManager) -> Tuple[mirpb.Annotation, str]:
    """
    generate mirpb.Annotation instance from object node in coco and pascal annotation xml file
    """
    name = _get_dom_xml_tag_data(obj, "name")
    bndbox_node = _get_dom_xml_tag_node(obj, "bndbox")
    xmin = int(float(_get_dom_xml_tag_data(bndbox_node, "xmin")))
    ymin = int(float(_get_dom_xml_tag_data(bndbox_node, "ymin")))
    xmax = int(float(_get_dom_xml_tag_data(bndbox_node, "xmax")))
    ymax = int(float(_get_dom_xml_tag_data(bndbox_node, "ymax")))
    width = xmax - xmin + 1
    height = ymax - ymin + 1

    annotation = mirpb.Annotation()
    annotation.class_id = class_type_manager.id_and_main_name_for_name(name)[0]
    annotation.box.x = xmin
    annotation.box.y = ymin
    annotation.box.w = width
    annotation.box.h = height
    annotation.score = 0
    return annotation


def import_annotations(mir_annotation: mirpb.MirAnnotations, mir_keywords: mirpb.MirKeywords, in_sha1_file: str,
                       annotations_dir_path: str, task_id: str):
    if (not in_sha1_file) or (not annotations_dir_path):
        logging.error("invalid input paths")
        return code.MirCode.RC_ERROR_INVALID_ARGS

    # read type_id_name_dict and type_name_id_dict
    class_type_manager = ids.ClassIdManager()
    logging.info("loaded type id and names: %d", class_type_manager.size())

    image_annotations = mir_annotation.task_annotations[task_id].image_annotations

    assethash_filename_list = []
    with open(in_sha1_file, "r") as in_file:
        for line in in_file.readlines():
            line_components = line.strip().split()
            if not line_components or len(line_components) < 2:
                logging.warning("incomplete line: %s", line)
                continue
            asset_hash, file_name = line_components[0], line_components[1]
            base_file_name = os.path.splitext(os.path.basename(file_name))[0]
            assethash_filename_list.append((asset_hash, base_file_name))

    logging.info("wrting %d annotations", len(assethash_filename_list))

    counter = 0
    cur_max_id = class_type_manager._max_id
    missing_annotations_counter = 0
    for asset_hash, base_file_name in assethash_filename_list:
        annotation_file = os.path.join(annotations_dir_path, base_file_name + '.xml')
        if not os.path.isfile(annotation_file):
            missing_annotations_counter += 1
            continue
        dom_tree = xml.dom.minidom.parse(annotation_file)
        if not dom_tree:
            logging.error("cannot open annotation_file: {}".format(annotation_file))
            return code.MirCode.RC_ERROR_INVALID_ARGS

        single_asset_keyids_set = set()
        collection = dom_tree.documentElement
        objects = collection.getElementsByTagName("object")
        for idx, obj in enumerate(objects):
            annotation = _xml_obj_to_annotation_and_name(obj, class_type_manager)
            annotation.index = idx
            image_annotations[asset_hash].annotations.append(annotation)
            single_asset_keyids_set.add(annotation.class_id)
        mir_keywords.keywords[asset_hash].predifined_keyids[:] = single_asset_keyids_set

        counter += 1
        if counter % 5000 == 0:
            logging.debug("finished %d / %d", counter, len(assethash_filename_list))

    # find extra class ids, save to new file and exit.
    max_id = class_type_manager._max_id
    if max_id > cur_max_id:
        new_ids = [class_type_manager._dirty_id_names_dict[i] for i in range(cur_max_id + 1, max_id + 1)]
        logging.error("\nError: import process abort because new class labels are detected: {}".format(
            ','.join(new_ids)))
        logging.error("new dict file is stored with a .new suffix at: {}".format(class_type_manager._csv_path))
        logging.error("contact ymir-team@intellif.com how to update ids file before continue,")
        logging.error("or use the .new file to replace the previous one a for temp running.")
        class_type_manager.save()
        return code.MirCode.RC_ERROR_INVALID_ARGS

    if missing_annotations_counter > 0:
        logging.warning(f"asset count that have no annotations: {missing_annotations_counter}")

    return code.MirCode.RC_OK



#    Modified on GitPython CMD wrapper:
#    https://github.com/gitpython-developers/GitPython/blob/master/git/cmd.py

import io
import logging
import os
import signal
from subprocess import (Popen, PIPE)
import subprocess
import sys
import threading
from collections import OrderedDict

from mir.scm.base import BaseScm


def dashify(string):
    return string.replace('_', '-')


def expand_path(pth, expand_vars=True):
    try:
        pth = os.path.expanduser(pth)
        if expand_vars:
            pth = os.path.expandvars(pth)
        return os.path.normpath(os.path.abspath(pth))
    except Exception:
        pass


def safe_decode(bin_str):
    """Safely decodes a binary string to unicode"""
    if isinstance(bin_str, str):
        return bin_str
    if isinstance(bin_str, bytes):
        return bin_str.decode(sys.getfilesystemencoding(), 'surrogateescape')
    if bin_str is not None:
        raise TypeError('Expected bytes or text, but got %r' % (bin_str, ))
    return None


def stream_copy(source, destination, chunk_size=512 * 1024):
    """Copy all data from the source stream into the destination stream in chunks
    of size chunk_size
    :return: amount of bytes written"""
    byte_read = 0
    while True:
        chunk = source.read(chunk_size)
        destination.write(chunk)
        byte_read += len(chunk)
        if len(chunk) < chunk_size:
            break
    # END reading output stream
    return byte_read


def find_root(root=None) -> str:
    root_dir = os.path.realpath(root or os.curdir)

    if not os.path.isdir(root_dir):
        raise ValueError("directory %s does not exist" % root_dir)

    return root_dir


def transform_kwarg(name, value, split_single_char_options):
    if len(name) == 1:
        if value is True:
            return ["-%s" % name]
        if value not in (False, None):
            if split_single_char_options:
                return ["-%s" % name, "%s" % value]
            return ["-%s%s" % (name, value)]
    else:
        if value is True:
            return ["--%s" % dashify(name)]
        if value is not False and value is not None:
            return ["--%s=%s" % (dashify(name), value)]
    return []


def transform_kwargs(split_single_char_options=True, **kwargs):
    """Transforms Python style kwargs into git command line options."""
    args = []
    kwargs = OrderedDict(sorted(kwargs.items(), key=lambda x: x[0]))
    for k, vals in kwargs.items():
        if isinstance(vals, (list, tuple)):
            for value in vals:
                args += transform_kwarg(k, value, split_single_char_options)
        else:
            args += transform_kwarg(k, vals, split_single_char_options)
    return args


execute_kwargs = {
    'istream', 'with_extended_output', 'with_exceptions', 'as_process', 'stdout_as_string', 'output_stream',
    'with_stdout', 'kill_after_timeout', 'universal_newlines', 'shell', 'env', 'max_chunk_size'
}


class CmdScm(BaseScm):
    """
    The Git class manages communication with the Git binary.
    It provides a convenient interface to calling the Git binary, such as in::
     g = Git( git_dir )
     g.init()                   # calls 'git init' program
     rval = g.ls_files()        # calls 'git ls-files' program
    ``Debugging``
        Set the GIT_PYTHON_TRACE environment variable print each invocation
        of the command to stdout.
        Set its value to 'full' to see details about the returned values.
    """
    def __init__(self, working_dir=None, scm_executable=None):
        """Initialize this instance with:
        :param working_dir:
           Git directory we should work in. If None, we always work in the current
           directory as returned by os.getcwd().
           It is meant to be the working tree directory if available, or the
           .git directory in case of bare repositories.
        :param scm_excutable: either git or dvc.
        """

        self._working_dir = find_root(working_dir)

        # self._working_dir = expand_path(working_dir)
        self._scm_executable = scm_executable

    def __getattr__(self, name):
        """A convenience method as it allows to call the command as if it was
        an object.
        :return: Callable object that will execute call _call_process with your arguments."""
        if name[0] == '_':
            return BaseScm.__getattr__(self, name)
        return lambda *args, **kwargs: self._call_process(name, *args, **kwargs)

    @property
    def working_dir(self):
        """:return: Git directory we are working on"""
        return self._working_dir

    def execute(self,
                command,
                istream=None,
                with_extended_output=False,
                with_exceptions=True,
                output_stream=None,
                stdout_as_string=True,
                kill_after_timeout=None,
                with_stdout=True,
                universal_newlines=False,
                env=None,
                max_chunk_size=io.DEFAULT_BUFFER_SIZE,
                **subprocess_kwargs):
        """Handles executing the command on the shell and consumes and returns
        the returned information (stdout)
        :param command:
            The command argument list to execute.
            It should be a string, or a sequence of program arguments. The
            program to execute is the first item in the args sequence or string.
        :param istream:
            Standard input filehandle passed to subprocess.Popen.
        :param with_extended_output:
            Whether to return a (status, stdout, stderr) tuple.
        :param with_exceptions:
            Whether to raise an exception when git returns a non-zero status.
        :param output_stream:
            If set to a file-like object, data produced by the git command will be
            output to the given stream directly. Processes will
            always be created with a pipe due to issues with subprocess.
            This merely is a workaround as data will be copied from the
            output pipe to the given output stream directly.
            Judging from the implementation, you shouldn't use this flag !
        :param stdout_as_string:
            if False, the commands standard output will be bytes. Otherwise, it will be
            decoded into a string using the default encoding (usually utf-8).
            The latter can fail, if the output contains binary data.
        :param env:
            A dictionary of environment variables to be passed to `subprocess.Popen`.
        :param max_chunk_size:
            Maximum number of bytes in one chunk of data passed to the output_stream in
            one invocation of write() method. If the given number is not positive then
            the default value is used.
        :param subprocess_kwargs:
            Keyword arguments to be passed to subprocess.Popen. Please note that
            some of the valid kwargs are already set by this method, the ones you
            specify may not be the same ones.
        :param with_stdout: If True, default True, we open stdout on the created process
        :param universal_newlines:
            if True, pipes will be opened as text, and lines are split at
            all known line endings.
        :param kill_after_timeout:
            To specify a timeout in seconds for the git command, after which the process
            should be killed. It is
            set to None by default and will let the process run until the timeout is
            explicitly specified. This feature is not supported on Windows. It's also worth
            noting that kill_after_timeout uses SIGKILL, which can have negative side
            effects on a repository. For example, stale locks in case of git gc could
            render the repository incapable of accepting changes until the lock is manually
            removed.
        :return:
            * str(output) if extended_output = False (Default)
            * tuple(int(status), str(stdout), str(stderr)) if extended_output = True
            if output_stream is True, the stdout value will be your output stream:
            * output_stream if extended_output = False
            * tuple(int(status), output_stream, str(stderr)) if extended_output = True
            Note git is executed with LC_MESSAGES="C" to ensure consistent
            output regardless of system language.
        :raise GitCommandError:
        :note:
           If you add additional keyword arguments to the signature of this method,
           you must update the execute_kwargs tuple housed in this module."""

        logging.debug("{}: {}".format(self.working_dir, ' '.join(command)))

        # Allow the user to have the command executed in their working dir.
        cwd = self._working_dir or os.getcwd()

        # Start the process
        inline_env = env
        env = os.environ.copy()
        # Attempt to force all output to plain ascii english, which is what some parsing code
        # may expect.
        # According to stackoverflow (http://goo.gl/l74GC8), we are setting LANGUAGE as well
        # just to be sure.
        env["LANGUAGE"] = "C"
        env["LC_ALL"] = "C"
        env.update()
        if inline_env is not None:
            env.update(inline_env)

        if sys.version_info[0] > 2:
            cmd_not_found_exception = FileNotFoundError  # NOQA # exists, flake8 unknown @UndefinedVariable
        else:
            cmd_not_found_exception = OSError
        # end handle

        stdout_sink = (PIPE if with_stdout else None)
        istream_ok = "None"
        if istream:
            istream_ok = "<valid stream>"
        try:
            proc = Popen(
                command,
                env=env,
                cwd=cwd,
                bufsize=-1,
                stdin=istream,
                stderr=PIPE,
                stdout=stdout_sink,
                close_fds=True,  # unsupported on windows
                universal_newlines=universal_newlines,
                creationflags=0,
                **subprocess_kwargs)
        except cmd_not_found_exception as err:
            raise ValueError(command, err) from err

        def _kill_process(pid):
            """ Callback method to kill a process. """
            process_id = Popen(['ps', '--ppid', str(pid)], stdout=PIPE)
            child_pids = []
            for line in process_id.stdout:
                if len(line.split()) > 0:
                    local_pid = (line.split())[0]
                    if local_pid.isdigit():
                        child_pids.append(int(local_pid))
            try:
                # Windows does not have SIGKILL, so use SIGTERM instead
                sig = getattr(signal, 'SIGKILL', signal.SIGTERM)
                os.kill(pid, sig)
                for child_pid in child_pids:
                    try:
                        os.kill(child_pid, sig)
                    except OSError:
                        pass
                kill_check.set()  # tell the main routine that the process was killed
            except OSError:
                # It is possible that the process gets completed in the duration after timeout
                # happens and before we try to kill the process.
                pass

        # end

        if kill_after_timeout:
            kill_check = threading.Event()
            watchdog = threading.Timer(kill_after_timeout, _kill_process, args=(proc.pid, ))

        # Wait for the process to return
        status = 0
        stdout_value = b''
        stderr_value = b''
        newline = "\n" if universal_newlines else b"\n"
        try:
            if output_stream is None:
                if kill_after_timeout:
                    watchdog.start()
                stdout_value, stderr_value = proc.communicate()
                if kill_after_timeout:
                    watchdog.cancel()
                    if kill_check.isSet():
                        stderr_value = ('Timeout: the command "%s" did not complete in %d '
                                        'secs.' % (" ".join(command), kill_after_timeout))
                        if not universal_newlines:
                            stderr_value = stderr_value.encode(sys.getfilesystemencoding())
                # strip trailing "\n"
                if stdout_value and stdout_value.endswith(newline):
                    stdout_value = stdout_value[:-1]
                if stderr_value and stderr_value.endswith(newline):
                    stderr_value = stderr_value[:-1]
                status = proc.returncode
            else:
                max_chunk_size = max_chunk_size if max_chunk_size and max_chunk_size > 0 else io.DEFAULT_BUFFER_SIZE
                stream_copy(proc.stdout, output_stream, max_chunk_size)
                stdout_value = proc.stdout.read()
                stderr_value = proc.stderr.read()
                # strip trailing "\n"
                if stderr_value.endswith(newline):
                    stderr_value = stderr_value[:-1]
                status = proc.wait()
            # END stdout handling
        finally:
            if proc.stdout:
                proc.stdout.close()
            if proc.stderr:
                proc.stderr.close()

        # # Keep this code incase we need to print debug infos.
        # if self.GIT_PYTHON_TRACE == 'full':
        #     cmdstr = " ".join(command)

        #     def as_text(stdout_value):
        #         return not output_stream and safe_decode(stdout_value) or '<OUTPUT_STREAM>'
        #     # end

        #     if stderr_value:
        #         logging.info("%s -> %d; stdout: '%s'; stderr: '%s'",
        #                  cmdstr, status, as_text(stdout_value), safe_decode(stderr_value))
        #     elif stdout_value:
        #         logging.info("%s -> %d; stdout: '%s'", cmdstr, status, as_text(stdout_value))
        #     else:
        #         logging.info("%s -> %d", cmdstr, status)
        # # END handle debug printing

        if with_exceptions and status != 0:
            raise ValueError(command, status, stderr_value, stdout_value)

        if isinstance(stdout_value, bytes) and stdout_as_string:  # could also be output_stream
            stdout_value = safe_decode(stdout_value)

        # Allow access to the command's status code
        if with_extended_output:
            return (status, stdout_value, safe_decode(stderr_value))
        return stdout_value

    @classmethod
    def __unpack_args(cls, arg_list):
        if not isinstance(arg_list, (list, tuple)):
            return [str(arg_list)]

        outlist = []
        for arg in arg_list:
            if isinstance(arg_list, (list, tuple)):
                outlist.extend(cls.__unpack_args(arg))
            # END recursion
            else:
                outlist.append(str(arg))
        # END for each arg
        return outlist

    def __call__(self, **kwargs):
        """Specify command line options to the git executable
        for a subcommand call
        :param kwargs:
            is a dict of keyword arguments.
            these arguments are passed as in _call_process
            but will be passed to the git command rather than
            the subcommand.
        ``Examples``::
            git(work_tree='/tmp').difftool()"""
        self._git_options = transform_kwargs(split_single_char_options=True, **kwargs)
        return self

    def _call_process(self, method, *args, **kwargs):
        """Run the given git command with the specified arguments and return
        the result as a String
        :param method:
            is the command. Contained "_" characters will be converted to dashes,
            such as in 'ls_files' to call 'ls-files'.
        :param args:
            is the list of arguments. If None is included, it will be pruned.
            This allows your commands to call git more conveniently as None
            is realized as non-existent
        :param kwargs:
            It contains key-values for the following:
            - the :meth:`execute()` kwds, as listed in :var:`execute_kwargs`;
            - "command options" to be converted by :meth:`transform_kwargs()`;
            - the `'insert_kwargs_after'` key which its value must match one of ``*args``,
              and any cmd-options will be appended after the matched arg.
        Examples::
            git.rev_list('master', max_count=10, header=True)
        turns into::
           git rev-list max-count 10 --header master
        :return: Same as ``execute``"""
        # Handle optional arguments prior to calling transform_kwargs
        # otherwise these'll end up in args, which is bad.
        exec_kwargs = {k: v for k, v in kwargs.items() if k in execute_kwargs}
        opts_kwargs = {k: v for k, v in kwargs.items() if k not in execute_kwargs}

        insert_after_this_arg = opts_kwargs.pop('insert_kwargs_after', None)

        # Prepare the argument list
        opt_args = transform_kwargs(**opts_kwargs)
        ext_args = self.__unpack_args([a for a in args if a is not None])

        if insert_after_this_arg is None:
            args = opt_args + ext_args
        else:
            try:
                index = ext_args.index(insert_after_this_arg)
            except ValueError as err:
                raise ValueError("Couldn't find argument '%s' in args %s to insert cmd options after" %
                                 (insert_after_this_arg, str(ext_args))) from err
            # end handle error
            args = ext_args[:index + 1] + opt_args + ext_args[index + 1:]
        # end handle opts_kwargs

        call = [self._scm_executable]

        call.append(dashify(method))
        call.extend(args)

        return self.execute(call, **exec_kwargs)



"""
use this module to read contents in other branch head ref or from other tags \n
some mir commands, such as `mir search`, `mir merge` will use this module
"""

import os
from typing import Any, Optional

import yaml

from mir import scm
from mir.tools import code


class ExodusError(Exception):
    __slots__ = ("code")

    def __init__(self, msg: str, code: int) -> None:
        super().__init__(msg)
        self.code = code


def locate_file_in_rev(mir_root: str, file_name: str, rev: str) -> str:
    """
    get the file location in mir_root for special rev
    Args:
    mir_root: root to a mir repo
    file_name: name of the mir file
    rev: branch name, commit id, or tag name
    """
    if not mir_root or not file_name or not rev:
        raise ValueError("invalid args")

    scm_git = scm.Scm(mir_root if mir_root else ".", scm_executable="git")

    # get `file_name`.dvc sha1
    # using git command: git rev-parse <rev>:<file_name>
    dvc_sha1 = scm_git.rev_parse("{}:{}.dvc".format(rev, file_name))
    if not dvc_sha1:
        raise ExodusError("found no dvc file: {}".format(file_name), code=code.MirCode.RC_ERROR_INVALID_MIR_FILE)

    # parse `file_name`.dvc to get file_name's sha1
    # using git command: git cat-file -p `dvc_sha1`
    # result should like:
    #   outs:
    #   - md5: ffa21df3e9741af03f60a60bb171e4a1
    #     size: 7304832
    #     path: keywords.mir
    # which is dvc format, get md5 and store it to `file_hash`
    dvc_file_str = scm_git.cat_file(["-p", dvc_sha1])
    dvc_file_yaml_data = yaml.safe_load(dvc_file_str)
    if not dvc_file_yaml_data.get("outs", None):
        raise ExodusError("found no singnature for file: {}".format(file_name),
                          code=code.MirCode.RC_ERROR_INVALID_MIR_FILE)

    file_hash = None  # type: Optional[str]
    out_list = dvc_file_yaml_data["outs"]
    for val in out_list:
        if val["path"] == file_name:
            file_hash = val["md5"]
            break

    if not file_hash:
        raise ExodusError("found no singnature for file: {}".format(file_name),
                          code=code.MirCode.RC_ERROR_INVALID_MIR_FILE)

    # open that file (it's in .dvc/cache) and return `file_name`
    file_path = os.path.join(mir_root, ".dvc/cache", file_hash[:2], file_hash[2:])
    return file_path


def _open_branch_tag_commit_file(mir_root: str, file_name: str, rev: str, mode: str) -> Any:
    """
    Opens file in mir repo from specific branch, tag or commit id
    Args:
        mir_root: root dir of mir repo
        file_name: mir file name
        rev: branch name, tag name or commit id
        mode: file open mode, same as in `open` function
    Returns:
        file descriptor or something
    Raises:
        io errors
    """
    if not rev:  # explict set rev, if use current rev, set to HEAD
        raise ExodusError("found no rev", code=code.MirCode.RC_ERROR_INVALID_ARGS)
    file_path = locate_file_in_rev(mir_root, file_name, rev)
    return open(file_path, mode)


class open_mir():
    __slots__ = ("_file_name", "_rev", "_mode", "_fd", "_mir_root")

    def __init__(self, mir_root: str, file_name: str, rev: str, mode: str):
        self._mir_root = mir_root
        self._file_name = file_name
        self._rev = rev
        self._mode = mode
        self._fd = None

    def __enter__(self):
        if not self._mir_root or not self._file_name or not self._mode:
            raise ValueError("invalid arguments")

        self._fd = _open_branch_tag_commit_file(mir_root=self._mir_root,
                                                file_name=self._file_name,
                                                rev=self._rev,
                                                mode=self._mode)
        return self._fd

    def __exit__(self, type, value, traceback):
        if self._fd is not None:
            self._fd.close()
            self._fd = None



from enum import IntEnum, auto


class MirCode(IntEnum):
    # everything is ok, command finished without any errors or warnings
    RC_OK = 0

    # errors: command failed for some reasons
    # error codes increase from 1
    RC_ERROR_UNKNOWN = auto()  # unknown error(s) occured while command executed
    RC_ERROR_INVALID_MIR_REPO = auto()  # mir command was not invoked inside a mir repo directory
    RC_ERROR_INVALID_ARGS = auto()  # lack of necessary command args, or unexpected operations detected
    RC_ERROR_INVALID_BRANCH_OR_TAG = auto()  # invalid branch name or tag name
    RC_ERROR_INVALID_MIR_FILE = auto()  # files like metadatas.mir, annotations.mir is not valid
    RC_ERROR_INVALID_COMMAND = auto()  # unknown command or sub command
    RC_ERROR_MIR_FILE_NOT_FOUND = auto()  # can not find some mir files
    RC_ERROR_CONFLICTS_OCCURED = auto()  # conflicts detected when mir pull or mir merge
    RC_ERROR_EMPTY_METADATAS = auto()  # no assets found in metadatas.mir
    RC_ERROR_EMPTY_TRAIN_SET = auto()  # no training set when training
    RC_ERROR_EMPTY_VAL_SET = auto()  # no validation set when training
    RC_ERROR_DIRTY_REPO = auto()  # no validation set when training
    RC_ERROR_NOTHING_TO_MERGE = auto()



"""mir command line interface"""
import argparse
import logging
import sys

from mir import version
from mir.commands import (init, branch, checkout, commit, exporting, filter, log, merge, reset, status, training, mining,
                          importing)

_COMMANDS_ = [init, branch, checkout, commit, exporting, filter, log, merge, reset, status, training, mining, importing]


class MirParser(argparse.ArgumentParser):
    """Custom parser class for mir CLI."""
    def error(self, message, cmd_cls=None):  # pylint: disable=arguments-differ
        logging.error(message)

    def parse_args(self, args=None, namespace=None):
        # NOTE: overriding to provide a more granular help message.
        args, argv = self.parse_known_args(args, namespace)
        if argv:
            msg = "unrecognized arguments: %s"
            self.error(msg % " ".join(argv), getattr(args, "func", None))
        return args


class VersionAction(argparse.Action):
    """Show mir version and exits"""
    def __call__(self, parser, namespace, values, option_string=None):
        logging.info("mir version: {0}, by Scalable AI team, Intellifusion".format(version.__version__))
        sys.exit(0)


class DebugModeAction(argparse.Action):
    """Work in debug mode, and print debug info."""
    def __call__(self, parser, namespace, values, option_string=None):
        for handler in logging.root.handlers[:]:
            logging.root.removeHandler(handler)

        logging.basicConfig(stream=sys.stdout,
                            format='%(levelname)-8s: [%(asctime)s] %(filename)s:%(lineno)s:%(funcName)s(): %(message)s',
                            datefmt='%Y%m%d-%H:%M:%S',
                            level=logging.DEBUG)
        logging.debug("in debug mode")


def create_main_parser() -> argparse.ArgumentParser:
    parent_parser = argparse.ArgumentParser()
    main_parser = MirParser(
        prog="mir",
        description="MIR",
        parents=[parent_parser],
        formatter_class=argparse.RawTextHelpFormatter,
        add_help=False,
    )

    main_parser.add_argument("-v", "--version", nargs=0, action=VersionAction, help="show program's version")

    main_parser.add_argument("-d", "--debug", nargs=0, action=DebugModeAction, help="set debug mode, print more infos.")

    share_parser = argparse.ArgumentParser(add_help=False)
    share_parser.add_argument("--root", "-r",
                              dest="mir_root",
                              type=str,
                              default=".",
                              help="root path to the mir repo, use . if not set.")

    # sub parsers
    subparsers = main_parser.add_subparsers(title="sub commands",
                                            metavar="COMMAND",
                                            dest="cmd",
                                            help="Use mir COMMAND --help for sub command help")

    for command in _COMMANDS_:
        command.bind_to_subparsers(subparsers, share_parser)

    return main_parser



from typing import List

from google.protobuf import message as pb_message

from ymir_proto.gen import mir_common_pb2 as mir_common
from ymir_proto.gen import mir_entities_pb2 as mirpb


def mir_type(mir_storage: mir_common.MirStorage):
    MIR_TYPE = {
        mir_common.MirStorage.MIR_METADATAS: mirpb.MirMetadatas,
        mir_common.MirStorage.MIR_ANNOTATIONS: mirpb.MirAnnotations,
        mir_common.MirStorage.MIR_KEYWORDS: mirpb.MirKeywords,
        mir_common.MirStorage.MIR_TASKS: mirpb.MirTasks,
    }
    return MIR_TYPE[mir_storage]


def mir_path(mir_storage: mir_common.MirStorage):
    MIR_PATH = {
        mir_common.MirStorage.MIR_METADATAS: 'metadatas.mir',
        mir_common.MirStorage.MIR_ANNOTATIONS: 'annotations.mir',
        mir_common.MirStorage.MIR_KEYWORDS: 'keywords.mir',
        mir_common.MirStorage.MIR_TASKS: 'tasks.mir',
    }
    return MIR_PATH[mir_storage]


def get_all_mir_paths():
    return [mir_path(ms) for ms in get_all_mir_storage()]


def get_all_mir_storage() -> List[pb_message.Message]:
    return [
        mir_common.MirStorage.MIR_METADATAS,
        mir_common.MirStorage.MIR_ANNOTATIONS,
        mir_common.MirStorage.MIR_KEYWORDS,
        mir_common.MirStorage.MIR_TASKS,
    ]



"""
exports the assets and annotations from mir format to ark-training-format
"""

from enum import Enum
import logging
import os
from typing import Dict, List, Optional, Set, Tuple
import xml.etree.ElementTree as ElementTree

from mir.tools import mir_storage_ops
from mir.tools import utils as mir_utils
from ymir_proto.gen import mir_common_pb2 as mir_common
from ymir_proto.gen import mir_entities_pb2 as mirpb
from ymir_proto.util import ids


class ExportError(Exception):
    """
    exception type raised by function `export`
    """
    pass


class ExportFormat(str, Enum):
    EXPORT_FORMAT_UNKNOWN = 'unknown'
    EXPORT_FORMAT_NO_ANNOTATION = 'none'
    EXPORT_FORMAT_ARK = 'ark'
    EXPORT_FORMAT_VOC = 'voc'


# public: format type
SUPPORTED_EXPORT_FORMATS = {
    ExportFormat.EXPORT_FORMAT_NO_ANNOTATION.value, ExportFormat.EXPORT_FORMAT_ARK.value,
    ExportFormat.EXPORT_FORMAT_VOC.value
}


def format_type_from_str(format: str) -> ExportFormat:
    return ExportFormat(format.lower())


# public: export
def export(mir_root: str, assets_location: str, class_type_ids: Set[int], asset_ids: str, asset_dir: str,
           annotation_dir: str, need_suffix: bool, base_branch: str,
           format_type: ExportFormat) -> bool:
    """
    export assets and annotations

    Args:
        mir_root (str): path to mir repo root directory
        assets_location (str): path to assets storage directory
        class_type_ids (Set[int]): class ids, all objects within this set will be exported, if None, export everything
        asset_ids (Set[str]): export asset ids
        asset_dir (str): asset directory
        annotation_dir (str): annotation directory, if format_type is NO_ANNOTATION, this could be None
        need_suffix (str): if true, all export assets will have it's type as suffix, jpg, png, etc.
        base_branch (str): data branch
        format_type (ExportFormat): format type, NONE means exports no annotations

    Raises:
        ValueError: if mir repo not provided

    Returns:
        True if export finished, raises errors when failed
    """
    if not mir_root:
        raise ValueError("invalid mir_repo")

    # export assets
    os.makedirs(asset_dir, exist_ok=True)
    asset_result = mir_utils.store_assets_to_dir(asset_ids=asset_ids,
                                                 out_root=asset_dir,
                                                 sub_folder=".",
                                                 asset_location=assets_location,
                                                 overwrite=False,
                                                 create_prefix=False,
                                                 need_suffix=need_suffix)

    logging.info(f"export {len(asset_result)} assets out of {len(asset_ids)}")

    if format_type == ExportFormat.EXPORT_FORMAT_NO_ANNOTATION:
        return True

    # export annotations
    mir_datas = mir_storage_ops.MirStorageOps.load(
        mir_root=mir_root,
        mir_branch=base_branch,
        mir_storages=[mir_common.MirStorage.MIR_METADATAS, mir_common.MirStorage.MIR_ANNOTATIONS])
    mir_annotations = mir_datas[mir_common.MirStorage.MIR_ANNOTATIONS]

    # add all annotations to assets_to_det_annotations_dict
    # key: asset_id as str, value: annotations as List[mirpb.Annotation]
    assets_to_det_annotations_dict = _annotations_by_assets(mir_annotations=mir_annotations,
                                                            class_type_ids=class_type_ids,
                                                            base_task_id=mir_annotations.head_task_id)

    if format_type == ExportFormat.EXPORT_FORMAT_ARK:
        _export_detect_ark_annotations_to_path(annotations_dict=assets_to_det_annotations_dict,
                                               asset_ids=asset_ids,
                                               dest_path=annotation_dir)
    elif format_type == ExportFormat.EXPORT_FORMAT_VOC:
        mir_metadatas = mir_datas[mir_common.MirStorage.MIR_METADATAS]
        _export_detect_voc_annotations_to_path(annotations_dict=assets_to_det_annotations_dict,
                                               mir_metadatas=mir_metadatas,
                                               asset_ids=asset_ids,
                                               dest_path=annotation_dir)
    else:
        raise ValueError(f"unsupported format: {format_type.name}")

    return True


def generate_asset_index_file(asset_dir: str, rel_start_dir: str, index_file_path: str,
                              image_suffixes: tuple = ('.jpg', '.jpeg', '.png')):
    """
    generate index file for export result.

    in index file, each line is a relpath start from dir or index_file_path, and points to export asset file

    examples:
        assets_dir is /tmp/assets, and rel_start_dir is /tmp \n
        then index file should be something like: \n
        assets/sha1sum-1.jpg \n
        assets/sha1sum-2.jpg

    Args:
        asset_dir (str): asset dir, every jpg / jpeg / png inside this dir will export to index.tsv
        rel_start_dir (str): start point of relative path, should exists
        index_file_path (str): path to index file you want to generate, should not exists
        image_suffixes (tuple): available image suffixes

    Raises:
        ValueError: if export_result, asset_dir or index_file_path empty
        ValueError: if index_file_path exists
        ValueError: if asset_id in export_result does not have an asset file name
    """
    if not asset_dir or not index_file_path or not image_suffixes:
        raise ValueError('invalid args')
    if not os.path.isdir(asset_dir):
        raise ValueError(f"invalid asset_dir: {asset_dir}, not a directory")
    if not os.path.isdir(rel_start_dir):
        raise ValueError(f"invalid rel_start_dir: {rel_start_dir}, not a directory")
    if os.path.exists(index_file_path):
        raise ValueError(f"index file already exists: {index_file_path}")

    index_dir = os.path.dirname(index_file_path)
    os.makedirs(index_dir, exist_ok=True)  # make sure that this dir exists

    relpath = os.path.relpath(asset_dir, start=rel_start_dir)

    with open(index_file_path, 'w') as idx_f:
        for item in os.listdir(asset_dir):
            if not os.path.isfile(os.path.join(asset_dir, item)) or os.path.splitext(item)[1] not in image_suffixes:
                continue
            idx_f.write(os.path.join(relpath, item) + '\n')


# private: export annotations: general
def _annotations_by_assets(mir_annotations: mirpb.MirAnnotations, class_type_ids: Set[int],
                           base_task_id: str) -> Dict[str, List[mirpb.Annotation]]:
    """
    get annotations by assets

    Args:
        mir_annotations (mirpb.MirAnnotations): annotations
        class_type_ids (Set[int]): only type ids within it could be output, if None, no class id filter applied
        base_task_id (str): base task id

    Returns:
        Dict, key: asset id, value: List[mirpb.Annotation]
    """
    assets_to_det_annotations_dict = {}  # type: Dict[str, List[mirpb.Annotation]]

    if base_task_id not in mir_annotations.task_annotations:
        raise ValueError(f"base task id: {base_task_id} not in mir_annotations")

    task_annotations = mir_annotations.task_annotations[base_task_id]
    for asset_id, image_annotations in task_annotations.image_annotations.items():
        matched_annotations = [
            annotation for annotation in image_annotations.annotations
            if (not class_type_ids or (annotation.class_id in class_type_ids))
        ]
        assets_to_det_annotations_dict[asset_id] = matched_annotations

    return assets_to_det_annotations_dict


# private: export annotations: ark
def _export_detect_ark_annotations_to_path(annotations_dict: Dict[str, List[mirpb.Annotation]], asset_ids: Set[str],
                                           dest_path: str) -> None:
    if not asset_ids:
        raise ValueError("empty asset ids")

    os.makedirs(dest_path, exist_ok=True)

    missing_counter = 0
    empty_counter = 0
    for asset_id in asset_ids:
        if asset_id not in annotations_dict:
            missing_counter += 1
            continue
        annotations = annotations_dict[asset_id]
        if len(annotations) == 0:
            empty_counter += 1
            continue

        file_name = f"{asset_id}.txt"
        dest_file_path = os.path.join(dest_path, file_name)
        _single_image_annotations_to_ark(annotations, dest_file_path)

    if missing_counter > 0 or empty_counter > 0:
        logging.warning(f"missing annotations assets: {missing_counter}, "
                        f"empty annotations assets: {empty_counter} out of {len(asset_ids)}")


# private: export annotations: voc
def _export_detect_voc_annotations_to_path(annotations_dict: Dict[str, List[mirpb.Annotation]],
                                           mir_metadatas: mirpb.MirMetadatas, asset_ids: Set[str],
                                           dest_path: str) -> None:
    if not asset_ids:
        raise ValueError('empty asset_ids')
    if not mir_metadatas:
        raise ValueError('invalid mir_metadatas')

    os.makedirs(dest_path, exist_ok=True)

    cls_id_mgr = ids.ClassIdManager()

    missing_counter = 0
    empty_counter = 0
    for asset_id in asset_ids:
        if asset_id not in mir_metadatas.attributes:
            raise ValueError(f"can not find asset id: {asset_id} in mir_metadatas")

        if asset_id not in annotations_dict:
            missing_counter += 1
            continue
        annotations = annotations_dict[asset_id]
        if len(annotations) == 0:
            empty_counter += 1
            continue

        attrs = mir_metadatas.attributes[asset_id]
        annotations = annotations_dict[asset_id]

        file_name = f"{asset_id}.xml"
        dest_file_path = os.path.join(dest_path, file_name)
        _single_image_annotations_to_voc(asset_id, attrs, annotations, dest_file_path, cls_id_mgr)

    if missing_counter > 0:
        logging.warning(f"missing annotations assets: {missing_counter}, empty annotations assets: {empty_counter}")


def _single_image_annotations_to_ark(annotations, dest_file_path):
    with open(dest_file_path, "w") as f:
        for annotation in annotations:
            f.write(f"{annotation.class_id}, {annotation.box.x}, {annotation.box.y}, "
                    f"{annotation.box.x + annotation.box.w - 1}, {annotation.box.y + annotation.box.h - 1}\n")


def _single_image_annotations_to_voc(asset_id, attrs, annotations, dest_file_path, cls_id_mgr):
    # annotation
    annotation_node = ElementTree.Element('annotation')

    # annotation: folder
    folder_node = ElementTree.SubElement(annotation_node, 'folder')
    folder_node.text = 'folder'

    # annotation: filename
    filename_node = ElementTree.SubElement(annotation_node, 'filename')
    filename_node.text = asset_id

    # annotation: source
    source_node = ElementTree.SubElement(annotation_node, 'source')

    # annotation: source: database
    database_node = ElementTree.SubElement(source_node, 'database')
    database_node.text = attrs.dataset_name or 'unknown'

    # annotation: source: annotation
    annotation2_node = ElementTree.SubElement(source_node, 'annotation')
    annotation2_node.text = 'unknown'

    # annotation: source: image
    image_node = ElementTree.SubElement(source_node, 'image')
    image_node.text = 'unknown'

    # annotation: size
    size_node = ElementTree.SubElement(annotation_node, 'size')

    # annotation: size: width
    width_node = ElementTree.SubElement(size_node, 'width')
    width_node.text = str(attrs.width)

    # annotation: size: height
    height_node = ElementTree.SubElement(size_node, 'height')
    height_node.text = str(attrs.height)

    # annotation: size: depth
    depth_node = ElementTree.SubElement(size_node, 'depth')
    depth_node.text = str(attrs.image_channels)

    # annotation: segmented
    segmented_node = ElementTree.SubElement(annotation_node, 'segmented')
    segmented_node.text = '0'

    # annotation: object(s)
    for annotation in annotations:
        object_node = ElementTree.SubElement(annotation_node, 'object')

        name_node = ElementTree.SubElement(object_node, 'name')
        name_node.text = cls_id_mgr.main_name_for_id(annotation.class_id) or 'unknown'

        pose_node = ElementTree.SubElement(object_node, 'pose')
        pose_node.text = 'unknown'

        truncated_node = ElementTree.SubElement(object_node, 'truncated')
        truncated_node.text = 'unknown'

        occluded_node = ElementTree.SubElement(object_node, 'occluded')
        occluded_node.text = '0'

        bndbox_node = ElementTree.SubElement(object_node, 'bndbox')

        xmin_node = ElementTree.SubElement(bndbox_node, 'xmin')
        xmin_node.text = str(annotation.box.x)

        ymin_node = ElementTree.SubElement(bndbox_node, 'ymin')
        ymin_node.text = str(annotation.box.y)

        xmax_node = ElementTree.SubElement(bndbox_node, 'xmax')
        xmax_node.text = str(annotation.box.x + annotation.box.w - 1)

        ymax_node = ElementTree.SubElement(bndbox_node, 'ymax')
        ymax_node.text = str(annotation.box.y + annotation.box.h - 1)

        difficult_node = ElementTree.SubElement(object_node, 'difficult')
        difficult_node.text = '0'

    # write to xml
    tree = ElementTree.ElementTree(annotation_node)
    tree.write(dest_file_path, encoding='utf-8')
